/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is ForumZilla 0.1.1.
 *
 * The Initial Developer of the Original Code is Zapogee, Inc.
 * Portions created by Zapogee, Inc. are Copyright (C) 2001
 * Zapogee, Inc.  All Rights Reserved.
 *
 * Contributor(s):
 *   Myk Melez, <myk@melez.com>, original author
 */


//****************************************************************************//
// global variables

var gFeedService = 
  Components
    .classes["@forumzilla.mozdev.org/feed/feedservice;1"]
      .getService(Components.interfaces.nsIFeedService);

var gMailSession;


//****************************************************************************//
// event handlers

/* load, unload, and selection handlers */

function doLoad() {
    updateModeButton();

    // Display the list of feed subscriptions.
    var ds = gFeedService.subscriptionsDS;
    var tree = document.getElementById('subscriptions');
    tree.database.AddDataSource(ds);
    tree.builder.rebuild();
    if (tree.view && tree.view.rowCount > 0)
        tree.view.selection.select(0);
    tree.focus();

    initializeFilterWidgets();

    // register the folder listener that listens for IMAP folder creation
    // and jumpstarts subfolder creation when importing hierarchical OPML.
    gMailSession =
      Components
        .classes["@mozilla.org/messenger/services/session;1"]
          .getService(Components.interfaces.nsIMsgMailSession);
    try {
        var nsIFolderListener = Components.interfaces.nsIFolderListener;
        var notifyFlags = nsIFolderListener.event;
        gMailSession.AddFolderListener(OPMLImportListener, notifyFlags);
    }
    catch (e) {
        debug("error adding folder listener to session: " + e);
    }
}

function doClose() {
    gMailSession.RemoveFolderListener(OPMLImportListener);
    window.close();
}

function doSelect(event) {
    var tree = document.getElementById('subscriptions');

    var titleField = document.getElementById('title');
    var urlField = document.getElementById('url');
    var folderField = document.getElementById('actionTargetFolder');
    var intervalField = document.getElementById('interval');
    var quickModeField = document.getElementById('quickmode');
    var enabledField = document.getElementById('enabled');
    var editCaption = document.getElementById('edit-caption');

    var i;
    var controls = ['title', 'url', 'actionTargetFolder', 'newFolderButton', 'interval', 'quickmode', 'enabled'];

    if (tree.view.selection.currentIndex == -1) {
        titleField.value = "";
        urlField.value = "";
        SetFolderPicker(gFeedService.storageFolder, 'actionTargetFolder');
        intervalField.value = "";
        quickModeField.selectedIndex = -1;
        enabledField.checked = false;
        editCaption.setAttribute('label', "Edit Feed");

        for ( i=0 ; i<controls.length ; i++ ) {
            document.getElementById(controls[i]).setAttribute('disabled', true);
        }
    }
    else {
        var item = tree.view.getItemAtIndex(tree.view.selection.currentIndex);
        var resource = rdf.GetResource(item.id);

        if (!item.feed)
            item.feed = gFeedService.getFeed(resource);

        for ( i=0 ; i<controls.length ; i++ ) {
            document.getElementById(controls[i]).removeAttribute('disabled');
        }

        titleField.value = item.feed.title;
        urlField.value = item.feed.url;
        SetFolderPicker(item.feed.folder.URI, 'actionTargetFolder');
        intervalField.value = item.feed.interval;
        quickModeField.selectedIndex = item.feed.quickMode ? 1 : 0;
        enabledField.checked = item.feed.enabled ? true : false;
        editCaption.setAttribute('label', "Edit " + item.feed.title);
    }
}



/* button handlers */

function doAdd() {
    var promptService =
      Components
        .classes["@mozilla.org/embedcomp/prompt-service;1"]
          .getService(Components.interfaces.nsIPromptService);
    var result = {value:null};
    var dummy = { value: 0 };
    if (promptService.prompt(null, "Add Feed", "Enter feed location:", result, null, dummy)
        && result.value.length)
        add(result.value);
}

function doRemove() {
    var tree = document.getElementById('subscriptions');
    var selectedIndex = tree.view.selection.currentIndex;
    var treeitem = tree.view.getItemAtIndex(selectedIndex);
    var feed = rdf.GetResource(treeitem.id);

    // Remove the feed from the list of subscriptions.
    var feeds = gFeedService.subscriptionsList;
    var index = feeds.IndexOf(feed);
    if (index != -1)
        feeds.RemoveElementAt(index, false);

    // Remove all assertions about the feed from the subscriptions database.
    var ds = gFeedService.subscriptionsDS;
    removeAssertions(ds, feed);

    // Remove all assertions about items in the feed from the items database.
    ds = gFeedService.itemsDS;
    var items = ds.GetSources(FZ_FEED, feed, true);
    while (items.hasMoreElements()) {
        var item = items.getNext();
        item = item.QueryInterface(Components.interfaces.nsIRDFResource);
        ds.Unassert(item, FZ_FEED, feed, true);
        if (ds.hasArcOut(item, FZ_FEED))
            debug(item.Value + " is from more than one feed; only the reference to this feed removed");
        else
            removeAssertions(ds, item);
    }

    tree.builder.rebuild();
    if (tree.view.rowCount > selectedIndex)
        tree.view.selection.select(selectedIndex);
    else if (tree.view.rowCount > 0)
        tree.view.selection.select(tree.view.rowCount-1);
    tree.focus();
}

function doCheck() {
    var tree = document.getElementById('subscriptions');
    downloadFeedAtIndex(tree.view.selection.currentIndex);
}

function toggleMode() {
    var advancedSection = document.getElementById('advanced-section');
    if (advancedSection.getAttribute('hidden') == 'true')
        advancedSection.setAttribute('hidden', 'false');
    else
        advancedSection.setAttribute('hidden', 'true');
    updateModeButton();
}

function doOpenPreferences() {
    window.openDialog('chrome://forumzilla/content/preferences.xul',
                      '',
                      'chrome,dialog,modal');
}

function doImport() {
    var rv = pickOpen('Pick OPML File to Import', '$xml $all');
    if (rv.reason == PICK_CANCEL)
        return false;

    var file = new LocalFile(rv.file, MODE_RDONLY);
    var text = file.read();
    var gParser = new DOMParser;
    var xmlDocument = gParser.parseFromString(text, 'text/xml');

    var outline = xmlDocument.getElementsByTagName("body")[0];
    var folderURI = nsPreferences.copyUnicharPref("forumzilla.storage.folder",
                                                  DEFAULT_STORAGE_FOLDER);
    var folder = GetMsgFolderFromUri(folderURI);
    //SetBusyCursor(window, true);
    try {
        importOutline(outline, folder);

        // If there are any keys in the queue, we're waiting on IMAP folder creation
        // to complete.  Otherwise, we're all done, so turn off the busy cursor.
        //for (var key in OPMLImportQueue) {
        //    debug("OPML import queue not empty; delaying turning off busy cursor");
        //    return true;
        //}
        //SetBusyCursor(window, false);
    }
    catch(e) {
        //SetBusyCursor(window, false);
        throw(e);
    }
    return true;
}

/* field update handlers */

function doUpdateURL() {
    var feed = getFeedForSelectedItem();
    feed.url = document.getElementById('url').value;
}

function doUpdateTitle() {
    var feed = getFeedForSelectedItem();
    feed.title = document.getElementById('title').value;
    document.getElementById('edit-caption').setAttribute('label', "Edit Feed: " + feed.title);
}

function doUpdateInterval() {
    var feed = getFeedForSelectedItem();
    feed.interval = document.getElementById('interval').value;
}

function doUpdateQuickMode() {
    var feed = getFeedForSelectedItem();
    feed.quickMode = document.getElementById('quickmode').value == "true" ? true : false;
}

function doUpdateEnabled() {
    var feed = getFeedForSelectedItem();
    feed.enabled = document.getElementById('enabled').checked;
}

function doUpdateFolder() {
    var feed = getFeedForSelectedItem();
    feed.folder = GetMsgFolderFromUri(document.getElementById('actionTargetFolder').getAttribute('uri'));

    // Rebuild the tree, since it doesn't seem to update otherwise;
    // maybe trees auto-rebuild for literals but not resources?
    var tree = document.getElementById('subscriptions');
    var index = tree.view.selection.currentIndex;
    tree.builder.rebuild();
    tree.view.selection.select(index);
}


//****************************************************************************//
// OPML import

// Arrays of outlines to be imported, indexed by folder ID, for outlines
// being imported into IMAP folders being created asynchronously.  We don't want
// to import the outline until the folder exists, so we queue it up here,
// then run the queue in the new folder listener.
//
// Note that while folder IDs may look like URIs, they aren't necessarily the same,
// since folder URIs may have truncated folder titles and additional characters
// like the single quotation mark (') encoded.
//
var OPMLImportQueue = {};

function importOutline(outline, folder) {
    // Recursively imports an OPML outline and its children.  Takes an outline
    // node from an OPML document and the folder into which the outline should
    // be imported.

    // Figure out if we're importing into an IMAP server and, if so, whether or not
    // it allows dual-use folders (folders containing both subfolders and messages).
    var imapServer;
    var dualUseFolders = true;
    try {
        if (folder.server)
            imapServer = folder.server.QueryInterface(Components.interfaces.nsIImapIncomingServer);
        if (imapServer)
            dualUseFolders = imapServer.dualUseFolders;
    }
    catch (e) {}

    // If this outline represents a feed, subscribe the user to the feed.
    var url = outline.getAttribute("xmlUrl");
    if (url) {
        var title = outline.getAttribute("title") || outline.getAttribute("text") || "untitled";
        if (outline.hasChildNodes() && !dualUseFolders) {
            // This outline represents a feed and a folder containing feeds,
            // but the server doesn't support dual-use folders, so we create
            // a subfolder within the folder for this outline and stick
            // the outline's feed into the subfolder.  To do that, we make
            // an item for the feed and add it to the outline's children.
            debug("dual-mode item " + title + " on single-use server; creating subfolder");
            var item = outline.ownerDocument.createElement("outline");
            item.setAttribute("xmlUrl", url);
            item.setAttribute("title", title);
            outline.insertBefore(item, outline.firstChild);
        }
        else
            add(url, title, folder.URI);
    }

    if (outline.hasChildNodes())
        importOutlineChildren(outline.childNodes, folder, imapServer, dualUseFolders);
}

function importOutlineChildren(items, parentFolder, imapServer, dualUseFolders) {
    for ( var i=0 ; i<items.length ; i++ ) {
        var item = items[i];

        // Only deal with "outline" elements; ignore text, etc. nodes.
        if (item.nodeName != "outline")
            continue;

        var title = item.getAttribute("title") || item.getAttribute("text") || "untitled";

        var storageMode = nsPreferences.getIntPref("forumzilla.storage.mode", DEFAULT_STORAGE_MODE);

        var childFolder;
        if (storageMode == FOLDER_STORAGE_MODE) {
            debug("folder mode; storing " + title + " in " + parentFolder.name);
            childFolder = parentFolder;
            importOutline(item, childFolder);
        }
        else if (storageMode == SUBFOLDERS_STORAGE_MODE) {
            debug("subfolders mode; storing " + title + " in " + parentFolder.name + "/" + title);
            var childFolderID = parentFolder.URI + "/" + title;
            if (parentFolder.containsChildNamed(title)) {
                debug(parentFolder.name + "/" + title + " already exists");
                childFolder = parentFolder.getChildNamed(title);
                childFolder = childFolder.QueryInterface(Components.interfaces.nsIMsgFolder);
                importOutline(item, childFolder);
            }
            else if (OPMLImportQueue[childFolderID]) {
                debug(parentFolder.name + "/" + title + " being created; queueing");
                OPMLImportQueue[childFolderID].push(item);
            }
            else {
                debug(parentFolder.name + "/" + title + " doesn't exist yet; creating");
                trailing = (item.hasChildNodes() && !dualUseFolders) ? "/" : "";
                parentFolder.createSubfolder(title + trailing, GetMsgWindow());
                if (imapServer) {
                    OPMLImportQueue[childFolderID] = [item];
                }
                else {
                    childFolder = parentFolder.getChildNamed(title);
                    childFolder = childFolder.QueryInterface(Components.interfaces.nsIMsgFolder);
                    importOutline(item, childFolder);
                }
            }
        }
    }
}

// The OPML import new folder listener.
var OPMLImportListener = {
    OnItemAdded: function(parentItem, item) {},
    OnItemRemoved: function(parentItem, item){},
    OnItemPropertyChanged: function(item, property, oldValue, newValue) {},
    OnItemIntPropertyChanged: function(item, property, oldValue, newValue) {},
    OnItemBoolPropertyChanged: function(item, property, oldValue, newValue) {},
    OnItemUnicharPropertyChanged: function(item, property, oldValue, newValue){},
    OnItemPropertyFlagChanged: function(item, property, oldFlag, newFlag) {},

    OnItemEvent: function(folder, event) {
        var eventType = event.toString();

        if (eventType != "FolderCreateCompleted" && eventType != "FolderCreateFailed")
            return;

        var childFolderID;
        if (folder && folder.parentMsgFolder) {
            childFolderID = folder.parentMsgFolder.URI + "/" + folder.name;
        }
        else {
            childFolderID = folder.URI;
        }

        var outlines = OPMLImportQueue[childFolderID];

        if (!outlines) {
            debug("no outlines for " + childFolderID);
            return;
        }
        else if (eventType == "FolderCreateCompleted") {
            debug(folder.URI + " create completed; importing outlines");
            for ( var i=0 ; i<outlines.length ; i++ ) {
                var outline = outlines[i];
                var title = outline.getAttribute('title') || outline.getAttribute('text');
                debug(folder.URI + " importing " + title);
                importOutline(outline, folder);
            }
        }
        else if (eventType == "FolderCreateFailed") {
            debug(folder.URI + " create failed");
        }

        delete OPMLImportQueue[childFolderID];

        // If there are any keys in the queue, we're waiting on IMAP folder creation
        // to complete.  Otherwise, we're all done, so turn off the busy cursor.
        //for (var key in OPMLImportQueue) {
        //    debug(childFolderID + " OPML import queue not empty; delaying turning off busy cursor");
        //    return;
        //}
        //debug("OPML import queue empty; turning off busy cursor");
        //SetBusyCursor(window, false);
    }
}


//****************************************************************************//
// new folder

function NewFolderOkCallback(childName, parentURI) {
    var parentFolder = GetMsgFolderFromUri(parentURI, true);

    // Figure out whether or not we're dealing with an IMAP folder.
    var imapFolder = null;
    try {
        imapFolder = parentFolder.QueryInterface(Components.interfaces.nsIMsgImapMailFolder);
    }
    catch(e) {}

    // IMAP folder creation is asynchronous, so if we're dealing with an IMAP
    // folder we have to listen for folder creation and update the picker only
    // after the IMAP service calls the listener.
    if (imapFolder) {
        if (!gMailSession)
            gMailSession =
              Components
                .classes["@mozilla.org/messenger/services/session;1"]
                  .getService(Components.interfaces.nsIMsgMailSession);
        try {
            var nsIFolderListener = Components.interfaces.nsIFolderListener;
            var notifyFlags = nsIFolderListener.event;
            if (gMailSession) {
                gMailSession.AddFolderListener(NewFolderListener, notifyFlags);
                SetBusyCursor(window, true);
            }
        }
        catch (e) {
            debug("error adding folder creation listener to session: " + e);
        }
    }

    // Create the folder.
    parentFolder.createSubfolder(childName, GetMsgWindow());
    
    // Set the folder picker to the newly created folder.
    if (!imapFolder) {
        var childFolderURI = parentFolder.getChildNamed(childName).URI;
        SetFolderPicker(childFolderURI, "actionTargetFolder");
    }
}

var NewFolderListener = {
    OnItemAdded: function(parentItem, item) {},
    OnItemRemoved: function(parentItem, item){},
    OnItemPropertyChanged: function(item, property, oldValue, newValue) {},
    OnItemIntPropertyChanged: function(item, property, oldValue, newValue) {},
    OnItemBoolPropertyChanged: function(item, property, oldValue, newValue) {},
    OnItemUnicharPropertyChanged: function(item, property, oldValue, newValue){},
    OnItemPropertyFlagChanged: function(item, property, oldFlag, newFlag) {},

    OnItemEvent: function(folder, event) {
        var eventType = event.toString();
        if (eventType == "FolderCreateCompleted") {
            SetFolderPicker(folder.URI, "actionTargetFolder");
            gMailSession.RemoveFolderListener(NewFolderListener);
            SetBusyCursor(window, false);
        }
        else if (eventType == "FolderCreateFailed") {
            SetBusyCursor(window, false);
        }
    }
}


//****************************************************************************//
// misc

function getFeedForSelectedItem() {
    var tree = document.getElementById('subscriptions');
    var item = tree.view.getItemAtIndex(tree.view.selection.currentIndex);
    if (item.feed)
      return item.feed;
    var feed =
      Components
        .classes["@forumzilla.mozdev.org/feed/feed;1"]
          .createInstance(Components.interfaces.nsIFeed);
    feed.init(rdf.GetResource(item.id));
    return feed;
}

function updateModeButton() {
    var advancedSection = document.getElementById('advanced-section');
    var modeButton = document.getElementById('mode-button');
    if (advancedSection.getAttribute('hidden') == 'true')
        modeButton.label = modeButton.getAttribute('advancedlabel');
    else
        modeButton.label = modeButton.getAttribute('simplelabel');
}

function downloadFeedAtIndex(index) {
    var tree = document.getElementById('subscriptions');
    var ds = gFeedService.subscriptionsDS;

    var item;
    try {
        item = tree.view.getItemAtIndex(index);
    } catch(e) {
        debug("couldn't get unselected item at index " + index);
        return;
    }
    var resource = rdf.GetResource(item.id);
    window.opener.gFeedService.downloadFeed(resource);
    //window.opener.gFeedService.storeItems();
}

function add(url, title, folder) {
    var tree = document.getElementById('subscriptions');
    var ds = gFeedService.subscriptionsDS;
    var feeds = gFeedService.subscriptionsList;

    // Generate a unique ID for the feed.
    var id = url;
    var i = 1;
    while (feeds.IndexOf(rdf.GetResource(id)) != -1 && ++i < 1000)
        id = url + i;
    if (id == 1000)
        throw("couldn't generate a unique ID for feed " + url);

    // Create the feed and add it to the database.
    var resource = rdf.GetResource(id);
    // XXX Perhaps the feed service should manage the subscriptions list
    // instead of us manually appending feeds to the list here.
    feeds.AppendElement(resource);
    ds.Assert(resource, RDF_TYPE, FZ_FEED, true);

    var feed = gFeedService.getFeed(resource);

    // Store information about the feed.
    feed.url = url;
    // In theory it isn't necessary to store this information for the feed,
    // since it'll pick up the default values, but the subscriptions tree won't
    // display the default values, so it's useful to add them. XXX Figure out
    // a better solution, since this prevents the values from updating
    // automatically when the defaults get updated.
    feed.quickMode = false;
    feed.interval = gFeedService.pollingInterval;
    feed.enabled = true;

    if (title)
        feed.title = title;
    else {
        // Download the feed to pick up the title.  Note that this downloads
        // the feed synchronously, which causes the UI to freeze while the feed
        // is being downloaded.  XXX We should figure out a better thing
        // to do here, at least tell the user to wait while we get this info,
        // but it would be even better to download it asynchronously and only
        // create the folder once we've received the title.
        feed.download(false, false);
    }

    if (folder)
        feed.folder = GetMsgFolderFromUri(folder);
    else
        // Access the folder to trigger the folder getter to create it.
        feed.folder;

    // Rebuild the tree, select the newly created feed, and focus the tree.
    tree.builder.rebuild();
    tree.view.selection.select(tree.view.rowCount-1);
    tree.focus();

    // Redownload the feed, getting the stories this time.
    feed.download(true, true);

    // Flush the data sources so we don't forget we've subscribed to this feed.
    // I'm not sure why this is necessary, but feeds seem to get forgotten otherwise.
    //gFeedService.flushDataSources();
}
